{*********************************************************}
{*                                                       *}
{*       TScrollButton 1.1 component                     *}
{*         put it on the form, assign Control property   *}
{*         and press MIDDLE mouse button on that control *}
{*                                                       *}
{*       Copyright (c) 1997 Alexander Petrosyan          *}
{*         email: paf@i-connect.ru, paf@fbit.msk.su      *}
{*         Fido : 2:5020/468.8                           *}
{*         Phone: (095)535-2222                          *}
{*                                                       *}
{*********************************************************}

unit ScrollBullet;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

const
  CURSOR_BASE          = 1973;
  CURSOR_DOWN          = CURSOR_BASE +  0;
  CURSOR_DOWN_LEFT     = CURSOR_BASE +  1;
  CURSOR_DOWN_RIGHT    = CURSOR_BASE +  2;
  CURSOR_FOUR_SIDES    = CURSOR_BASE +  3;
  CURSOR_LEFT          = CURSOR_BASE +  4;
  CURSOR_RIGHT         = CURSOR_BASE +  5;
  CURSOR_UP            = CURSOR_BASE +  6;
  CURSOR_UP_DOWN       = CURSOR_BASE +  7;
  CURSOR_UP_LEFT       = CURSOR_BASE +  8;
  CURSOR_UP_RIGHT      = CURSOR_BASE +  9;
  CURSOR_LEFT_RIGHT    = CURSOR_BASE + 10;

const
  DefaultScrollSpeedDivider = 50;

type
  TScrollBullet = class(TCustomControl)
  private
    FControl: TWinControl;
    FSpeedDivider: Integer;
  private
    procedure SetEvents;
    procedure RestoreEvents;
  private
    SavedMouseDown: TMouseEvent;
    procedure ControlMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    AboveBullet: Boolean;
    SB_V, SB_H: Integer;
    BULLET_: PChar;
    procedure Activate;
    procedure Deactivate;
    procedure Update(X, Y: Integer);
    procedure KillTimers;
    procedure WMTIMER(var Msg: TWMTimer);
      message WM_TIMER;
  protected
    procedure Paint; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Control: TWinControl read FControl write FControl;
    property Width stored False;
    property Height stored False;
    property SpeedDivider: Integer read FSpeedDivider write FSpeedDivider default DefaultScrollSpeedDivider;
  end;

procedure Register;

implementation

{$R *}

const
  Size = 32-1;
  Size2 = Size div 2;
const
  SB_NONE = -1;

type
  _TWinControl = class(TWinControl);  // Hack to get to TWinControl.OnMouseDown event

constructor TScrollBullet.Create;
begin
  inherited Create(AOwner);

  Width := Size;
  Height := Size;
  Visible := False;

  FSpeedDivider := DefaultScrollSpeedDivider;
end;

destructor TScrollBullet.Destroy;
begin
  RestoreEvents;

  inherited Destroy;
end;

procedure TScrollBullet.Loaded;
begin
  inherited Loaded;

  if not (csDesigning in ComponentState) then
  begin
    SetWindowRgn(Handle, CreateEllipticRgn(0, 0, Size, Size), False);
    SetEvents;
  end;
end;

procedure TScrollBullet.Paint;
var
  R: TRect;
  Bitmap: TBitmap;
begin
  R := ClientRect;
  with Canvas do
  begin
    if csDesigning in ComponentState then
      BULLET_ := 'BULLET_FOUR_SIDES';

    Bitmap := TBitmap.Create;
    try
      Bitmap.Handle := LoadBitmap(hInstance, BULLET_);
      Draw(0, 0, Bitmap);
    finally
      Bitmap.Free;
    end;

    if not (csDesigning in ComponentState) then
    begin
      Pen.Color := clBlack;
      Pen.Width := 3;
      Brush.Style := bsClear;
      Ellipse(0, 0, Size, Size);
    end;
  end;
end;

procedure TScrollBullet.Notification;
begin
  if AComponent = FControl then
    FControl := nil;
end;

procedure TScrollBullet.SetEvents;
begin
  if Assigned(FControl) then
    with _TWinControl(FControl) do
    begin
      SavedMouseDown := OnMouseDown;
      OnMouseDown := ControlMouseDown;
    end;
end;

procedure TScrollBullet.RestoreEvents;
begin
  if Assigned(FControl) then
    _TWinControl(FControl).OnMouseDown := SavedMouseDown;
end;

procedure TScrollBullet.ControlMouseDown;
begin
  if Button = mbMiddle
  then
    Activate
  else
    if Assigned(SavedMouseDown) then
      SavedMouseDown(Sender, Button, Shift, X, Y);
end;

procedure TScrollBullet.MouseUp;
begin
  inherited MouseUp(Button, Shift, X, Y);

  if not AboveBullet then
    Deactivate;
end;

procedure TScrollBullet.MouseDown;
begin
  inherited MouseDown(Button, Shift, X, Y);

  Deactivate;
end;

procedure TScrollBullet.Activate;
var
  P: TPoint;
begin
  GetCursorPos(P);
  P := Parent.ScreenToClient(P);

  Left := P.X - Size2;
  Top := P.Y - Size2;
  Update(0, 0);

  Show;
  MouseCapture := True;
end;

procedure TScrollBullet.KillTimers;
begin
  KillTimer(Handle, 1);
end;

procedure TScrollBullet.Deactivate;
begin
  KillTimers;

  Screen.Cursor := crDefault;
  MouseCapture := False;
  Hide;
end;

procedure TScrollBullet.Update;
type
  TCompare = (Less, Equal, More);

  function CompareWithBullet(Mouse: Integer): TCompare;
  begin
    if Mouse < 0
    then Result := Less
    else
      if Mouse > Size
      then Result := More
      else Result := Equal;
  end;

  function GetCursor(CX, CY: TCompare): TCursor;
  const
    Cursors: array[TCompare, TCompare] of TCursor = (
        (CURSOR_UP_LEFT, CURSOR_UP, CURSOR_UP_RIGHT),
        (CURSOR_LEFT, 0, CURSOR_RIGHT),
        (CURSOR_DOWN_LEFT, CURSOR_DOWN, CURSOR_DOWN_RIGHT)
      );
  begin
    Result := Cursors[CY, CX];
  end;

  const
    SB_Array: array[TCompare{Y}, TCompare{X}] of Integer = (
      (SB_LINEUP, SB_LINEUP, SB_LINEUP),
      (SB_NONE, SB_NONE, SB_NONE),
      (SB_LINEDOWN, SB_LINEDOWN, SB_LINEDOWN)
    );

  BULLET_ARRAY: array[Boolean{X}, Boolean{Y}] of PChar = (
    ('BULLET_UP_DOWN'{when nowere to scroll}, 'BULLET_UP_DOWN'),
    ('BULLET_LEFT_RIGHT', 'BULLET_FOUR_SIDES')
  );
  CURSOR_ARRAY: array[Boolean{X}, Boolean{Y}] of TCursor = (
    (CURSOR_UP_DOWN{when nowere to scroll}, CURSOR_UP_DOWN),
    (CURSOR_LEFT_RIGHT, CURSOR_FOUR_SIDES)
  );

var
  CX, CY: TCompare;
  MinPos, MaxPos: Integer;
  CanScrollX, CanScrollY: Boolean;
  Cursor: TCursor;
  Distance: Integer;
begin
  CX := CompareWithBullet(X);
  CY := CompareWithBullet(Y);

  GetScrollRange(Control.Handle, SB_HORZ, MinPos, MaxPos); CanScrollX := MinPos <> MaxPos;
  GetScrollRange(Control.Handle, SB_VERT, MinPos, MaxPos); CanScrollY := MinPos <> MaxPos;
  if not CanScrollX then CX := Equal;
  if not CanScrollY then CY := Equal;

  BULLET_ := BULLET_ARRAY[CanScrollX, CanScrollY];

  Cursor := GetCursor(CX, CY);
  AboveBullet := Cursor = 0;
  if AboveBullet then
    Cursor := CURSOR_ARRAY[CanScrollX, CanScrollY];
  Screen.Cursor := Cursor;

  SB_V := SB_Array[CY, CX];
  SB_H := SB_Array[CX, CY];

  KillTimers;
  Distance :=
    Abs(X div Size2 - 1)+
    Abs(Y div Size2 - 1);
  if Distance <> 0 then
    SetTimer(Handle, 1, SpeedDivider div Distance, nil);
end;

procedure TScrollBullet.MouseMove;
begin
  inherited MouseMove(Shift, X, Y);

  Update(X, Y);
end;

procedure TScrollBullet.WMTIMER;
begin
  with Control do
  begin
    if SB_V <> SB_NONE then Perform(WM_VSCROLL, SB_V, 0);
    if SB_H <> SB_NONE then Perform(WM_HSCROLL, SB_H, 0);
  end;
end;

{ LoadCursors }

procedure LoadCursors;
type
  TCursorEntry = record
    Value: TCursor;
    Name: PChar;
  end;
const
  Cursors: array[0..10] of TCursorEntry = (
    (Value: CURSOR_FOUR_SIDES;     Name: 'CURSOR_FOUR_SIDES'),
    (Value: CURSOR_UP_DOWN;        Name: 'CURSOR_UP_DOWN'),
    (Value: CURSOR_LEFT_RIGHT;     Name: 'CURSOR_LEFT_RIGHT'),

    (Value: CURSOR_UP;             Name: 'CURSOR_UP'),
    (Value: CURSOR_UP_RIGHT;       Name: 'CURSOR_UP_RIGHT'),
    (Value: CURSOR_RIGHT;          Name: 'CURSOR_RIGHT'),
    (Value: CURSOR_DOWN_RIGHT;     Name: 'CURSOR_DOWN_RIGHT'),
    (Value: CURSOR_DOWN;           Name: 'CURSOR_DOWN'),
    (Value: CURSOR_DOWN_LEFT;      Name: 'CURSOR_DOWN_LEFT'),
    (Value: CURSOR_LEFT;           Name: 'CURSOR_LEFT'),
    (Value: CURSOR_UP_LEFT;        Name: 'CURSOR_UP_LEFT'));
var
  I: Integer;
begin
  for I := Low(Cursors) to High(Cursors) do
    with Cursors[I] do
      Screen.Cursors[Value] := LoadCursor(hInstance, Name);
end;


{ Register }

procedure Register;
begin
  RegisterComponents('PAF', [TScrollBullet]);
end;


initialization
  LoadCursors;
end.
